# This is a slightly modified copy of https://github.com/holgerteichgraeber/kshape/tree/master/kshape/core.py
# Special thanks to https://github.com/Mic92/kshape for providing the initial code for the implementation of the kshape method

import sys
import math
import numpy as np

from numpy.random import randint
from numpy.linalg import norm, eigh
from numpy.fft import fft, ifft
import pdb
from sklearn.externals.joblib import Parallel,delayed
from sklearn.utils import check_random_state

def zscore(a, axis=0, ddof=0):
    a = np.asanyarray(a)
    mns = a.mean(axis=axis)
    sstd = a.std(axis=axis, ddof=ddof)
    if axis and mns.ndim < a.ndim:
        res = ((a - np.expand_dims(mns, axis=axis)) /
               np.expand_dims(sstd, axis=axis))
    else:
        res = (a - mns) / sstd
    return np.nan_to_num(res)


def roll_zeropad(a, shift, axis=None):
    a = np.asanyarray(a)
    if shift == 0:
        return a
    if axis is None:
        n = a.size
        reshape = True
    else:
        n = a.shape[axis]
        reshape = False
    if np.abs(shift) > n:
        res = np.zeros_like(a)
    elif shift < 0:
        shift += n
        zeros = np.zeros_like(a.take(np.arange(n-shift), axis))
        res = np.concatenate((a.take(np.arange(n-shift, n), axis), zeros), axis)
    else:
        zeros = np.zeros_like(a.take(np.arange(n-shift, n), axis))
        res = np.concatenate((zeros, a.take(np.arange(n-shift), axis)), axis)
    if reshape:
        return res.reshape(a.shape)
    else:
        return res


def _ncc_c(x, y):
    """
    >>> _ncc_c([1,2,3,4], [1,2,3,4])
    array([ 0.13333333,  0.36666667,  0.66666667,  1.        ,  0.66666667,
            0.36666667,  0.13333333])
    >>> _ncc_c([1,1,1], [1,1,1])
    array([ 0.33333333,  0.66666667,  1.        ,  0.66666667,  0.33333333])
    >>> _ncc_c([1,2,3], [-1,-1,-1])
    array([-0.15430335, -0.46291005, -0.9258201 , -0.77151675, -0.46291005])
    """
    den = np.array(norm(x) * norm(y))
    den[den == 0] = np.Inf

    x_len = len(x)
    fft_size = 1 << (2*x_len-1).bit_length()
    cc = ifft(fft(x, fft_size) * np.conj(fft(y, fft_size)))
    cc = np.concatenate((cc[-(x_len-1):], cc[:x_len]))
    return np.real(cc) / den


def _sbd(x, y):
    """
    >>> _sbd([1,1,1], [1,1,1])
    (-2.2204460492503131e-16, array([1, 1, 1]))
    >>> _sbd([0,1,2], [1,2,3])
    (0.043817112532485103, array([1, 2, 3]))
    >>> _sbd([1,2,3], [0,1,2])
    (0.043817112532485103, array([0, 1, 2]))
    """
    ncc = _ncc_c(x, y)
    idx = ncc.argmax()
    dist = 1 - ncc[idx]
    yshift = roll_zeropad(y, (idx + 1) - max(len(x), len(y)))

    return dist, yshift


def _extract_shape(idx, x, j, cur_center, normalize=True):
    """
    >>> _extract_shape(np.array([0,1,2]), np.array([[1,2,3], [4,5,6]]), 1, np.array([0,3,4]))
    array([-1.,  0.,  1.])
    >>> _extract_shape(np.array([0,1,2]), np.array([[-1,2,3], [4,-5,6]]), 1, np.array([0,3,4]))
    array([-0.96836405,  1.02888681, -0.06052275])
    >>> _extract_shape(np.array([1,0,1,0]), np.array([[1,2,3,4], [0,1,2,3], [-1,1,-1,1], [1,2,2,3]]), 0, np.array([0,0,0,0]))
    array([-1.2089303 , -0.19618238,  0.19618238,  1.2089303 ])
    >>> _extract_shape(np.array([0,0,1,0]), np.array([[1,2,3,4],[0,1,2,3],[-1,1,-1,1],[1,2,2,3]]), 0, np.array([-1.2089303,-0.19618238,0.19618238,1.2089303]))
    array([-1.19623139, -0.26273649,  0.26273649,  1.19623139])
    """
    _a = []
    for i in range(len(idx)):
        if idx[i] == j:
            if cur_center.sum() == 0:
                opt_x = x[i]
            else:
                _, opt_x = _sbd(cur_center, x[i])
            _a.append(opt_x)
    a = np.array(_a)

    if len(a) == 0:
        return np.zeros((1, x.shape[1]))
    columns = a.shape[1]
    if normalize:
        y = zscore(a, axis=1, ddof=1)
        s = np.dot(y.transpose(), y)

        p = np.empty((columns, columns))
        p.fill(1.0/columns)
        p = np.eye(columns) - p

        m = np.dot(np.dot(p, s), p)
        _, vec = eigh(m)
    else:  
        s = np.dot(a.transpose(), a)
        m = s  
        _, vec = eigh(m)
   
    centroid = vec[:, -1]
    finddistance1 = math.sqrt(((a[0] - centroid) ** 2).sum())
    finddistance2 = math.sqrt(((a[0] + centroid) ** 2).sum())

    if finddistance1 >= finddistance2:
        centroid *= -1
 
    return zscore(centroid, ddof=1) if normalize else centroid


def _kshape(x, k, n_init=1, max_iter=100, n_jobs = 1, random_state=None,normalize=True ):
    """
    >>> from numpy.random import seed; seed(0)
    >>> _kshape(np.array([[1,2,3,4], [0,1,2,3], [-1,1,-1,1], [1,2,2,3]]), 2)
    (array([0, 0, 1, 0]), array([[-1.2244258 , -0.35015476,  0.52411628,  1.05046429],
           [-0.8660254 ,  0.8660254 , -0.8660254 ,  0.8660254 ]]))
    """
    #print "n jobs run in parallel: " + str(cpu_count() ) 
    random_state = check_random_state(random_state)
    best_tot_dist,best_centroids,best_idx = None,None,None
    
    if n_jobs ==1:

        for i_init in range(n_init):    
            # n_init is the number of random starting points
            # pdb.set_trace()
            
            idx, centroids,tot_dist = _kshape_single(x, k, max_iter=max_iter, random_state= random_state,normalize=normalize) 
            if best_tot_dist is None or tot_dist < best_tot_dist:
                best_idx = idx.copy()
                best_centroids = centroids.copy()
                best_tot_dist = tot_dist
    else: # n_jobs not =1 # if -1, all CPUs are used
        # parallelisation of kshape runs
        seeds = random_state.randint(np.iinfo(np.int32).max,size=n_init)
        results = Parallel(n_jobs=n_jobs, verbose=0)(
            delayed(_kshape_single)(x,k,max_iter=max_iter, random_state=seed, normalize=normalize)
            for seed in seeds )
        # Get results with the lowest distances
        idx, centroids,tot_dist, iterations = zip(*results)
        best = np.argmin(tot_dist) 
        best_idx = idx[best]
        best_centroids = centroids[best]
        best_tot_dist = tot_dist[best]
    sys.stdout.write("Done: k="+str(k)+"\n")
    return {'centroids':best_centroids, 'labels':best_idx, 'distance':best_tot_dist,'centroids_all':centroids,'labels_all':idx,'distance_all':tot_dist,'iterations':iterations}

def _kshape_single(x, k, max_iter=10000, random_state=None,normalize=True):

    random_state = check_random_state(random_state)
    m = x.shape[0] # number of data points (365)

    empty_clust = True
    while empty_clust:  # make sure no empty cluster is generated  
        idx = random_state.randint(0, k, size=m)
        if np.unique(idx).shape[0] == k:
            empty_clust = False
    centroids = np.zeros((k, x.shape[1]))
    distances = np.empty((m, k))

    for _ in range(max_iter):
        # increase iterations with each k
        old_idx = idx
        if np.unique(idx).shape[0] != k: # if one of the clusters is empty, start anew
            empty_clust = True
            while empty_clust:  # make sure no empty cluster is generated  
                idx = random_state.randint(0, k, size=m)
                if np.unique(idx).shape[0] == k:
                    empty_clust = False
        for j in range(k):
            centroids[j] = _extract_shape(idx, x, j, centroids[j],normalize=normalize)

        for i in range(m):
            for j in range(k):
                distances[i, j] = 1 - max(_ncc_c(x[i], centroids[j]))
        idx = distances.argmin(1)

        sys.stdout.flush() # empty the buffer
        if np.array_equal(old_idx, idx):
            sys.stdout.write( "iter: " +  str(_) + " k=" + str(k)+"\n")
            sys.stdout.flush() # empty the buffer
            iterations = _
            break
        elif _==(max_iter-1):
            sys.stdout.write( "iter: " + str(_) + " k=" + str(k)+"\n")
            sys.stdout.flush() # empty the buffer 
            iterations = _
    dist_daily = distances.min(1)
    tot_dist = np.sum(dist_daily) # total distance, for n_init>1

    return idx, centroids, tot_dist, iterations


def kshape(x, k):
    idx, centroids = _kshape(np.array(x), k)
    clusters = []
    for i, centroid in enumerate(centroids):
        series = []
        for j, val in enumerate(idx):
            if i == val:
                series.append(j)
        clusters.append((centroid, series))
    return clusters

if __name__ == "__main__":
    import sys
    import doctest
    sys.exit(doctest.testmod()[0])
